function [symbols, initProb, states, tranProb, emitProb, err]=nma_readHMM(fileName)
%function [symbols, initProb, states, tranProb, emissionProb]=nma_readHMM(fileName)
%
% This function read the HMM configuration file for my final project
% for MATH 127, UC Berkeley. Fall 2002.
%
% INPUT:
%   fileName: the full path of the configuration file.
%
% OUTPUT:
%   symbols: a CELL array of characters. The allowed observation symbols.
%
%   initProb: an array of numbers, there will be as many number as there
%             are states
%             
%   states: a cell array of strings. The names of the states.
%
%   tranProb: a matrix, nxn, where n is the number of states. the order
%             of states is that indicated by the states array.
%
%   emissionProb: The probability of emission of a symbol from each state.
%                 an nxm matrix, where n is the number of states, and m
%                 is the number of symbols. The order of the states is as
%                 indicated by the states array, and the order of the 
%
%                 symbols is as indicated by the symbols array.
%  err: a string, if empty, indicates no error in parsing the file, else
%                 the string shows the parsing error.
%

% by Nasser Abbasi

err='';
symbols={''};
initProb=[];
states={''};
tranProb=[];
emitProb=[];

[fid,err]=fopen(fileName,'rt');
if(fid==-1)
   return;
end

TRUE=1;
FALSE=0;

INITIAL=0;
STATES=1;
statesFound=FALSE;

INIT_PROB=2;
initProbFound=FALSE;

SYMBOLS=3;
symbolsFound=FALSE;

EMIT_PROB=4;
emitProbFound=FALSE;

TRAN_PROB=5;
tranProbFound=FALSE;

DONE=6;

nState=0;
nInitProb=0;
nSym=0;
nEmitProbStates=0;
nTranProbLines=0;

currentState=INITIAL;

while(currentState ~=DONE)
   switch(currentState)
      case DONE
         if(fid ~= -1)
            fclose(fid);
         end
         return;
         
      case INITIAL
         %fprintf('initial state\n');
         err=findToken('<states>',fid);
         if(isempty(err))
            currentState=STATES;
         else
            err=['failed to find states section' err];
            currentState=DONE;
         end
         
         
      case STATES
         % looking for list of names, one per line
         %fprintf('states state\n');
         line=fgetl(fid);
         while(line ~= -1)
            if(length(line)==0)
               line=fgetl(fid)
            else
               break;
            end            
         end
         
         if(line==-1)
            err='Invalid HMM file. incomplete states section';
            currentState=DONE;
         else            
            line=nma_trim(line);
            if(length(line)>0)             
               if(line(1) ~='#')
                  if(isequal(line,'<init_prob>'))
                     if(nState==0)
                        err='No states found in the states section.';
                        currentState=DONE;
                     else                        
                        currentState=INIT_PROB;
                        statesFound=TRUE;
                     end
                  else
                     nState=nState+1;
                     states(nState)={line};
                  end
               end
            end
         end
         
      case INIT_PROB
         %looking for name=value, one per line. each name must be
         %allready found in the STATES.
         %fprintf('init_prob state\n');
         line=fgetl(fid);
         while(line ~= -1)
            if(length(line)==0)
               line=fgetl(fid)
            else
               break;
            end            
         end
         
         if(line==-1)
            err='Invalid HMM file. incomplete init_prob section';
            currentState=DONE;
         else            
            line=nma_trim(line);
            if(length(line)>0)             
               if(line(1) ~='#')
                  if(isequal(line,'<symbols>'))
                     if(nInitProb==0)
                        err='No init_prob entries found in the init_prob section.';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        return;                        
                     end
                     
                     if(nInitProb ~= nState)
                        err='Number of init_prob entries must equal number of states.';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        return;                        
                     end
                     
                     currentState=SYMBOLS;
                     initProbFound=TRUE;
                  else
                     %parse value  from line. allready trimmed.
                     i=0; 
                     thisValue=0;
                     
                     thisValue=str2double(line);
                     if(thisValue==NaN)
                        err=sprintf('invalid numeric value for probability in init_prob section. Found %s',line);
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        return;                        
                     end
                     
                     nInitProb=nInitProb+1;
                     if(nInitProb>nState)
                        err='Too many entries in init_prob. should equal to number of states';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        return;                        
                     end
                     
                     initProb(nInitProb)=thisValue;
                  end
               end
            end
         end
         
      case SYMBOLS
         %parse line with this a,b,c,d format
         %fprintf('symbols state\n');
         
         line=fgetl(fid);
         while(line ~= -1)
            if(length(line)==0)
               line=fgetl(fid)
            else
               break;
            end            
         end
         
         if(line==-1)
            err='Invalid HMM file. incomplete symbols section';
            return;
         end
         
         line=nma_trim(line);
         if(length(line)>0)             
            if(line(1) ~='#')
               if(isequal(line,'<emit_prob>'))
                  if(nSym==0)
                     err='No symbols found in the emit_prob section.';
                     if(fid ~= -1)
                        fclose(fid);
                     end
                     
                     return;
                  end
                  
                  currentState=EMIT_PROB;
               else
                  lookForComma=FALSE;                        
                  %parse values from line. allready trimmed.
                  if(length(line)==1)
                     nSym=nSym+1;
                     symbols(nSym)={line};
                  else                                                       
                     i=0;                        
                     while(1)
                        i=i+1;
                        if(i>length(line))
                           break;
                        end
                        
                        while(i<=length(line) & (line(i)==' ' | line(i)=='\t' ))
                           i=i+1;
                        end
                        
                        if(i>length(line))
                           break;
                        end
                        
                        if(lookForComma)
                           if(line(i)~=',')
                              err='expected a comma in symbols section, did not find it';
                              if(fid ~= -1)
                                 fclose(fid);
                              end
                              
                              return;
                           else
                              lookForComma=FALSE;
                           end
                        end                                                                    
                        
                        if(line(i) ~=',')                                    
                           nSym=nSym+1;
                           symbols(nSym)={line(i)};
                           lookForComma=TRUE;
                        end                               
                     end                                 
                  end                       
               end                    
            end
         end
         
      case EMIT_PROB
         %parse value,value,value 
         %fprintf('emission_prob state\n');
         
         line=fgetl(fid);
         while(line ~= -1)
            if(length(line)==0)
               line=fgetl(fid)
            else
               break;
            end            
         end
         
         if(line==-1)
            err='Invalid HMM file. incomplete emit_prob section';
            if(fid ~= -1)
               fclose(fid);
            end
            
            return;
         end
         
         line=nma_trim(line);
         if(length(line)>0)             
            if(line(1) ~='#')
               if(isequal(line,'<tran_prob>'))
                  if(nEmitProbStates==0)
                     err='No emit_prob entries found in the emit_prob section.';
                     if(fid ~= -1)
                        fclose(fid);
                     end
                     
                     return;
                  end
                  
                  if(nEmitProbStates ~= nState)
                     err='number of lines in emit_prob section must equal to number of states';
                     return;
                  end
                  
                  currentState=TRAN_PROB;
               else
                  %parse values from line. allready trimmed.
                  
                  lookForComma=FALSE;
                  nValuesFoundOnThisLine=0;
                  values=[];
                  i=1;
                  while(1)
                     if(line(i)~=',' & (line(i)==' ' | line(i)=='\t'))
                        i=i+1;
                        if(i>length(line))
                           break;
                        end
                     end
                     
                     if(lookForComma)
                        if(line(i)~=',')
                           err='expected a comma in emit_prob section';
                           if(fid ~= -1)
                              fclose(fid);
                           end
                           
                           return;
                        else
                           lookForComma=FALSE;
                           i=i+1;
                        end                                   
                     end
                     
                     if(i>length(line))
                        break;
                     end
                     
                     while(line(i)==' ' | line(i)=='\t')
                        i=i+1;
                        if(i>length(line))
                           break;
                        end
                     end
                     
                     startPos=i;
                     while(i<length(line) & line(i)~=' ' &line(i)~='\t' &line(i)~=',')
                        i=i+1;
                     end
                     
                     if(i==startPos)
                        endPos=i;
                     else
                        if(line(i)==' ' | line(i)=='\t' | line(i)==',')
                           endPos=i-1;
                        else
                           endPos=i;
                        end
                     end
                     
                     value=line(startPos:endPos);
                     value=str2double(value);
                     if(value==NaN)
                        err='Invalid numeric value for emit_prob found';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        
                        return;
                     end
                     
                     nValuesFoundOnThisLine=nValuesFoundOnThisLine+1;
                     
                     if(nValuesFoundOnThisLine>nSym)
                        err='emit probability line contains too many entries, more than number of symbols';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        
                        return;
                     end
                     
                     values(nValuesFoundOnThisLine)=value;
                     if(i==length(line))
                        if(nValuesFoundOnThisLine ~= nSym)
                           err='emit_prob section, missing emit_probability for some symbols';
                           if(fid ~= -1)
                              fclose(fid);
                           end
                           
                           return;
                        end
                        nEmitProbStates=nEmitProbStates+1;
                        if(nEmitProbStates>nState)
                           err='number of lines in emit_prob section can not be more than the number of states';
                           if(fid ~= -1)
                              fclose(fid);
                           end
                           
                           return;
                        end
                        emitProb(nEmitProbStates,1:nSym)=values(1:end);
                        break;
                     else
                        lookForComma=TRUE;
                     end
                  end
               end
            end
         end           
         
      case TRAN_PROB
         %parse matrix  v11,v12
         %              v21,v22
         %there will be n number of lines, where n is the number of
         %states
         
         %fprintf('tran_prob state\n');
         
         line=fgetl(fid);
         
         while(line ~= -1)
            if(length(line)==0)
               line=fgetl(fid)
            else
               break;
            end            
         end
         
         if(line==-1)            
            if(nTranProbLines < nState)
               err='Invalid HMM file. incomplete tran_prob section';
            end
            
            if(fid ~= -1)
               fclose(fid);
            end              
            
            return;
         end
         
         line=nma_trim(line);
         if(length(line)>0)             
            if(line(1) ~='#')   
               if(nTranProbLines==nState)
                  err='Too many lines in tran_prob section. can only have same lines as number of states';
                  if(fid ~= -1)
                     fclose(fid);
                  end              
                  
                  return;
                  
               end
               
               lookForComma=FALSE;
               nValuesFoundOnThisLine=0;
               values=[];
               i=1;
               while(1)
                  if(line(i)~=',' & (line(i)==' ' | line(i)=='\t'))
                     i=i+1;
                     if(i>length(line))
                        break;
                     end
                  end
                  
                  if(lookForComma)
                     if(line(i)~=',')
                        err='expected a comma in tran_prob section';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        
                        return;
                     else
                        lookForComma=FALSE;
                        i=i+1;
                     end                                   
                  end
                  
                  if(i>length(line))
                     break;
                  end
                  
                  while(line(i)==' ' | line(i)=='\t')
                     i=i+1;
                     if(i>length(line))
                        break;
                     end
                  end
                  
                  startPos=i;
                  while(i<length(line) & line(i)~=' ' &line(i)~='\t' &line(i)~=',')
                     i=i+1;
                  end
                  
                  if(i==startPos)
                     endPos=i;
                  else
                     if(line(i)==' ' | line(i)=='\t' | line(i)==',')
                        endPos=i-1;
                     else
                        endPos=i;
                     end
                  end
                  
                  value=line(startPos:endPos);
                  value=str2double(value);
                  if(value==NaN)
                     err='Invalid numeric value for tran_prob found';
                     if(fid ~= -1)
                        fclose(fid);
                     end
                     
                     return;
                  end
                  
                  nValuesFoundOnThisLine=nValuesFoundOnThisLine+1;
                  
                  if(nValuesFoundOnThisLine>nState)
                     err='tran probability line contains too many entries, more than number of states';
                     if(fid ~= -1)
                        fclose(fid);
                     end
                     
                     return;
                  end
                  
                  values(nValuesFoundOnThisLine)=value;
                  if(i==length(line))
                     if(nValuesFoundOnThisLine ~= nState)
                        err='tran_prob section, missing tran_probability for some states';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        
                        return;
                     end
                     nTranProbLines=nTranProbLines+1;
                     if(nTranProbLines>nState)
                        err='number of lines in tran_prob section can not be more than the number of states';
                        if(fid ~= -1)
                           fclose(fid);
                        end
                        
                        return;
                     end
                     tranProb(nTranProbLines,1:nState)=values(1:end);
                     break;
                  else
                     lookForComma=TRUE;
                  end
               end                        
            end            
         end
   end    
end    



%%%%%%%%%%%%%%%
%
%
%%%%%%%%%%%%%%%%%%%%
function err=findToken(token,fid)

err='';
found=0;

while(~found)
   
   line=fgetl(fid);
   while(line ~= -1)
      if(length(line)==0)
         line=fgetl(fid)
      else
         break;
      end            
   end
   
   if(line==-1)
      err='Invalid HMM file. Expected <states> section, did not find it';
      return;
   end
   
   line=nma_trim(line);
   if(length(line)>0)             
      if(line(1) ~='#')
         if(isequal(line,token))
            %fprintf('found %s\n',token);
            found=1;
         else
            err='Failed to find <states> section where it is expected';
         end
      end
   end
end




